# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fsspec
from fsspec.registry import known_implementations
from tos import EnvCredentialsProvider

from tosfs import TosFileSystem
from tosfs.utils import random_str


def test_fssepc_register():
    fsspec.register_implementation("tos", "tosfs.TosFileSystem")

    tosfs, _ = fsspec.core.url_to_fs(
        "tos://",
        endpoint=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
    )
    assert len(tosfs.ls("")) > 0


def test_set_known_implementations():
    known_implementations["tos"] = {"class": "tosfs.core.TosFileSystem"}

    tosfs, _ = fsspec.core.url_to_fs(
        "tos://",
        endpoint_url=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
    )
    assert len(tosfs.ls("")) > 0


def test_fsspec_open(bucket, temporary_workspace):
    known_implementations["tos"] = {"class": "tosfs.core.TosFileSystem"}

    file = f"{random_str()}.txt"
    content = "Hello TOSFS."
    with fsspec.open(
        f"tos://{bucket}/{temporary_workspace}/{file}",
        endpoint=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
        mode="w",
    ) as f:
        f.write(content)

    with fsspec.open(
        f"tos://{bucket}/{temporary_workspace}/{file}",
        endpoint_url=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
        mode="r",
    ) as f:
        assert f.read() == content


class MyTosFileSystem(TosFileSystem):
    """A sub class of TosFileSystem."""

    def __init__(self):
        """Init MyTosFileSystem."""
        super().__init__(
            endpoint=os.environ.get("TOS_ENDPOINT"),
            region=os.environ.get("TOS_REGION"),
            credentials_provider=EnvCredentialsProvider(),
        )


def test_customized_class(bucket, temporary_workspace):
    fsspec.register_implementation("tos", MyTosFileSystem, True)

    file = f"{random_str()}.txt"
    content = "Hello TOSFS."

    with fsspec.open(f"tos://{bucket}/{temporary_workspace}/{file}", mode="w") as f:
        f.write(content)

    with fsspec.open(f"tos://{bucket}/{temporary_workspace}/{file}", mode="r") as f:
        assert f.read() == content



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from tosfs.core import TosFileSystem
from tosfs.utils import find_bucket_key


@pytest.mark.parametrize(
    ("input_str", "expected_output"),
    [
        ("bucket/key", ("bucket", "key")),
        ("bucket/", ("bucket", "")),
        ("tos://bucket/key", ("bucket", "key")),
        ("tos://bucket/", ("bucket", "")),
        ("/key", ("", "key")),
        ("bucket/key/with/slashes", ("bucket", "key/with/slashes")),
        ("bucket", ("bucket", "")),
        ("", ("", "")),
    ],
)
def test_find_bucket_key(
    tosfs: TosFileSystem, input_str: str, expected_output: str
) -> None:
    assert find_bucket_key(input_str) == expected_output



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import tempfile
from typing import Any

import pytest

from tosfs.utils import create_temp_dir, random_str


def test_copy(fsspecfs: Any, bucket: str, temporary_workspace: str):
    # Create a temporary directory and files
    dir_name = random_str()
    subdir_path = f"{bucket}/{temporary_workspace}/{dir_name}"
    fsspecfs.mkdir(subdir_path)
    file1_path = f"{subdir_path}/file1.txt"
    file2_path = f"{subdir_path}/file2.txt"

    with fsspecfs.open(file1_path, "wb") as f:
        f.write(b"Hello, World!")
    with fsspecfs.open(file2_path, "wb") as f:
        f.write(b"Goodbye, World!")

    # Test Case 1: Copy single file
    copy_dest = f"{bucket}/{temporary_workspace}/copy_file1.txt"
    fsspecfs.copy(file1_path, copy_dest)
    assert fsspecfs.exists(copy_dest), "Failed to copy single file"
    assert (
        fsspecfs.cat_file(copy_dest) == b"Hello, World!"
    ), "Content mismatch in copied file"

    # Test Case 2: Copy directory recursively
    copy_dir_dest = f"{bucket}/{temporary_workspace}/copy_dir"
    fsspecfs.mkdir(copy_dir_dest)
    fsspecfs.copy(subdir_path.rstrip("/") + "/", copy_dir_dest, recursive=True)
    assert fsspecfs.exists(
        f"{copy_dir_dest}/file1.txt"
    ), "Failed to copy directory recursively"
    assert fsspecfs.exists(
        f"{copy_dir_dest}/file2.txt"
    ), "Failed to copy directory recursively"
    assert (
        fsspecfs.cat_file(f"{copy_dir_dest}/file1.txt") == b"Hello, World!"
    ), "Content mismatch in copied directory"
    assert (
        fsspecfs.cat_file(f"{copy_dir_dest}/file2.txt") == b"Goodbye, World!"
    ), "Content mismatch in copied directory"

    # Test Case 3: Copy non-existent file with on_error="ignore"
    non_existent_path = f"{subdir_path}/nonexistent.txt"
    copy_non_existent_dest = f"{bucket}/{temporary_workspace}/copy_nonexistent.txt"
    fsspecfs.copy(non_existent_path, copy_non_existent_dest, on_error="ignore")
    assert not fsspecfs.exists(
        copy_non_existent_dest
    ), "Non-existent file should not be copied"

    # Test Case 4: Copy non-existent file with on_error="raise"
    with pytest.raises(FileNotFoundError):
        fsspecfs.copy(non_existent_path, copy_non_existent_dest, on_error="raise")

    # Test Case 5: Copy file when dest/target exists
    copy_dest = f"{bucket}/{temporary_workspace}/copy_file1.txt"
    assert fsspecfs.exists(copy_dest), "Destination file should exist"
    assert fsspecfs.cat_file(copy_dest) == b"Hello, World!", "Content mismatch"
    fsspecfs.copy(file2_path, copy_dest)
    assert fsspecfs.exists(copy_dest), "Failed to copy and overwrite existing file"
    assert (
        fsspecfs.cat_file(copy_dest) == b"Goodbye, World!"
    ), "Content mismatch in overwritten file"


def test_info(fsspecfs: Any, bucket: str, temporary_workspace: str):
    # Setup
    file_path = f"{bucket}/{temporary_workspace}/test_file.txt"
    dir_path = f"{bucket}/{temporary_workspace}/test_dir"
    nested_file_path = f"{dir_path}/nested_file.txt"
    fsspecfs.mkdir(dir_path)
    with fsspecfs.open(file_path, "wb") as f:
        f.write(b"Hello, World!")
    with fsspecfs.open(nested_file_path, "wb") as f:
        f.write(b"Nested Content")

    # Test file info
    file_info = fsspecfs.info(file_path)
    assert file_info["name"] == fsspecfs._strip_protocol(
        file_path
    ), "Incorrect file name"
    expected_file_size = 13
    assert file_info["size"] == expected_file_size, "Incorrect file size"
    assert file_info["type"] == "file", "Incorrect type for file"

    # Test directory info
    dir_info = fsspecfs.info(dir_path)
    assert dir_info["name"] == fsspecfs._strip_protocol(
        dir_path
    ), "Incorrect directory name"
    assert dir_info["type"] == "directory", "Incorrect type for directory"
    # Some FS might not support 'size' for directories, so it's not strictly checked

    # Test non-existent path
    with pytest.raises(FileNotFoundError):
        fsspecfs.info(f"{bucket}/{temporary_workspace}/non_existent")

    # Test protocol stripping
    protocol_included_path = fsspecfs._strip_protocol(file_path)
    protocol_info = fsspecfs.info(protocol_included_path)
    assert protocol_info["name"] == protocol_included_path, "Protocol stripping failed"


def test_write_and_read_bytes(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path_to_write = f"{bucket}/{temporary_workspace}/{file_name}.bin"
    data_to_write = b"Hello, World!"

    with fsspecfs.open(path_to_write, "wb") as f:
        f.write(data_to_write)

    assert fsspecfs.exists(path_to_write), f"Path {path_to_write} does not exist."

    with fsspecfs.open(path_to_write, "rb") as f:
        data_to_read = f.read()

    assert (
        data_to_read == data_to_write
    ), f"Data read from {path_to_write} is not the same as data written."


def test_write_and_read_text(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path_to_write = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data_to_write = "Hello, World!"

    with fsspecfs.open(path_to_write, "w") as f:
        f.write(data_to_write)

    assert fsspecfs.exists(path_to_write), f"Path {path_to_write} does not exist."

    with fsspecfs.open(path_to_write, "r") as f:
        data_to_read = f.read()

    assert (
        data_to_read == data_to_write
    ), f"Data read from {path_to_write} is not the same as data written."


def test_with_size(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"

    expected_file_size = 10

    data = b"a" * (expected_file_size * 1**10)

    with fsspecfs.open(path, "wb") as f:
        f.write(data)

    with fsspecfs.open(path, "rb", size=10) as f:
        assert f.size == expected_file_size
        out = f.read()
        assert len(out) == expected_file_size


def test_simple(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data = b"a" * (10 * 1**10)

    with fsspecfs.open(path, "wb") as f:
        f.write(data)

    with fsspecfs.open(path, "rb") as f:
        out = f.read(len(data))
        assert len(data) == len(out)
        assert out == data


def test_write_large(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    mb = 2**20
    payload_size = int(2.5 * 1 * mb)
    payload = b"0" * payload_size

    with fsspecfs.open(path, "wb") as fd:
        fd.write(payload)

    assert fsspecfs.info(path)["size"] == payload_size


def test_write_limit(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    mb = 2**20
    block_size = 1 * mb
    payload_size = 2 * mb
    payload = b"0" * payload_size

    with fsspecfs.open(path, "wb", blocksize=block_size) as fd:
        fd.write(payload)

    assert fsspecfs.info(path)["size"] == payload_size


def test_readline(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"

    lines_to_write = [b"First line\n", b"Second line\n", b"Third line"]

    with fsspecfs.open(path, "wb") as f:
        for line in lines_to_write:
            f.write(line)

    with fsspecfs.open(path, "rb") as f:
        for expected_line in lines_to_write:
            read_line = f.readline()
            assert (
                read_line == expected_line
            ), f"Expected {expected_line}, got {read_line}"

        assert f.readline() == b"", "Expected empty string when reading past the end"


def test_readline_empty(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data = b""
    with fsspecfs.open(path, "wb") as f:
        f.write(data)
    with fsspecfs.open(path, "rb") as f:
        result = f.readline()
        assert result == data


def test_readline_blocksize(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data = b"ab\n" + b"a" * (1 * 2**20) + b"\nab"
    with fsspecfs.open(path, "wb") as f:
        f.write(data)
    with fsspecfs.open(path, "rb") as f:
        result = f.readline()
        expected = b"ab\n"
        assert result == expected

        result = f.readline()
        expected = b"a" * (1 * 2**20) + b"\n"
        assert result == expected

        result = f.readline()
        expected = b"ab"
        assert result == expected


def test_next(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.csv"

    csv_content = b"name,amount,id\nAlice,100,1\nBob,200,2\nCharlie,300,3\n"

    with fsspecfs.open(path, "wb") as f:
        f.write(csv_content)

    with fsspecfs.open(path, "rb") as f:
        # Step 5: Use __next__ to read the first line and assert
        expected = csv_content.split(b"\n")[0] + b"\n"
        result = next(f)
        assert result == expected, f"Expected {expected}, got {result}"

    fsspecfs.rm(path)


def test_iterable(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}"
    data = b"abc\n123"
    with fsspecfs.open(path, "wb") as f:
        f.write(data)
    with fsspecfs.open(path) as f, io.BytesIO(data) as g:
        for froms3, fromio in zip(f, g):
            assert froms3 == fromio
        f.seek(0)
        assert f.readline() == b"abc\n"
        assert f.readline() == b"123"
        f.seek(1)
        assert f.readline() == b"bc\n"

    with fsspecfs.open(path) as f:
        out = list(f)
    with fsspecfs.open(path) as f:
        out2 = f.readlines()
    assert out == out2
    assert b"".join(out) == data


def test_write_read_without_protocol(
    fsspecfs: Any, bucket: str, temporary_workspace: str
):
    file_name = random_str()
    path_to_write = f"{bucket}/{temporary_workspace}/{file_name}.bin"
    path_without_protocol = fsspecfs._strip_protocol(path_to_write)
    data_to_write = b"Hello, World!"

    with fsspecfs.open(path_without_protocol, "wb") as f:
        f.write(data_to_write)

    assert fsspecfs.exists(
        path_without_protocol
    ), f"Path {path_without_protocol} does not exist."

    with fsspecfs.open(f"/{path_without_protocol}", "rb") as f:
        data_to_read = f.read()

    assert (
        data_to_read == data_to_write
    ), f"Data read from {path_to_write} is not the same as data written."


def test_walk(fsspecfs: Any, bucket: str, temporary_workspace: str):
    nested_dir_1 = f"{bucket}/{temporary_workspace}/nested_dir_1"
    nested_dir_2 = f"{nested_dir_1}/nested_dir_2"
    file_1 = f"{bucket}/{temporary_workspace}/file_1.txt"
    file_2 = f"{nested_dir_1}/file_2.txt"
    file_3 = f"{nested_dir_2}/file_3.txt"

    fsspecfs.mkdir(nested_dir_1)
    fsspecfs.mkdir(nested_dir_2)
    with fsspecfs.open(file_1, "wb") as f:
        f.write(b"File 1 content")
    with fsspecfs.open(file_2, "wb") as f:
        f.write(b"File 2 content")
    with fsspecfs.open(file_3, "wb") as f:
        f.write(b"File 3 content")

    # Test walk with maxdepth=None and topdown=True
    result = list(
        fsspecfs.walk(f"{bucket}/{temporary_workspace}", maxdepth=None, topdown=True)
    )
    expected = [
        (
            fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}"),
            ["nested_dir_1"],
            ["file_1.txt"],
        ),
        (fsspecfs._strip_protocol(nested_dir_1), ["nested_dir_2"], ["file_2.txt"]),
        (fsspecfs._strip_protocol(nested_dir_2), [], ["file_3.txt"]),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test walk with maxdepth=1 and topdown=True
    result = list(
        fsspecfs.walk(f"{bucket}/{temporary_workspace}", maxdepth=1, topdown=True)
    )
    expected = [
        (
            fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}"),
            ["nested_dir_1"],
            ["file_1.txt"],
        ),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test walk with maxdepth=2 and topdown=True
    result = list(
        fsspecfs.walk(f"{bucket}/{temporary_workspace}", maxdepth=2, topdown=True)
    )
    expected = [
        (
            fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}"),
            ["nested_dir_1"],
            ["file_1.txt"],
        ),
        (fsspecfs._strip_protocol(nested_dir_1), ["nested_dir_2"], ["file_2.txt"]),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test walk with maxdepth=None and topdown=False
    result = list(
        fsspecfs.walk(f"{bucket}/{temporary_workspace}", maxdepth=None, topdown=False)
    )
    expected = [
        (fsspecfs._strip_protocol(nested_dir_2), [], ["file_3.txt"]),
        (fsspecfs._strip_protocol(nested_dir_1), ["nested_dir_2"], ["file_2.txt"]),
        (
            fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}"),
            ["nested_dir_1"],
            ["file_1.txt"],
        ),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test walk with detail=True
    result = list(
        fsspecfs.walk(
            f"{bucket}/{temporary_workspace}", maxdepth=None, topdown=True, detail=True
        )
    )
    expected_dir_num = 3
    assert (
        len(result) == expected_dir_num
    ), f"Expected {expected_dir_num} directories, got {len(result)}"
    for _, dirs, files in result:
        assert isinstance(dirs, dict), f"Expected dirs to be dict, got {type(dirs)}"
        assert isinstance(files, dict), f"Expected files to be dict, got {type(files)}"


def test_find(fsspecfs: Any, bucket: str, temporary_workspace: str):
    def remove_last_modification_time_ms(data):
        if isinstance(data, dict):
            for key in data:
                if "LastModified" in data[key]:
                    del data[key]["LastModified"]
        return data

    file1_path = f"{bucket}/{temporary_workspace}/file1"
    file2_path = f"{bucket}/{temporary_workspace}/file2"
    dir1_path = f"{bucket}/{temporary_workspace}/dir1"
    dir2_path = f"{bucket}/{temporary_workspace}/dir2"
    file3_path = f"{dir1_path}/file3"
    file4_path = f"{dir2_path}/file4"

    fsspecfs.mkdir(dir1_path)
    fsspecfs.mkdir(dir2_path)

    fsspecfs.touch(file1_path)
    fsspecfs.touch(file2_path)
    fsspecfs.touch(file3_path)
    fsspecfs.touch(file4_path)

    # Test finding all files
    result = fsspecfs.find(f"{bucket}/{temporary_workspace}")
    expected = [
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir1/file3"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir2/file4"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test finding files with maxdepth=1
    result = fsspecfs.find(f"{bucket}/{temporary_workspace}", maxdepth=1)
    expected = [
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"),
    ]
    assert result == expected, f"Expected {expected}, got {result}"

    # Test finding files and directories
    result = fsspecfs.find(f"{bucket}/{temporary_workspace}", withdirs=True)
    expected = [
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir1"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir1/file3"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir2"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir2/file4"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"),
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"),
    ]
    assert sorted(result) == sorted(expected), f"Expected {expected}, got {result}"

    # Test finding files with detail=True
    result = fsspecfs.find(f"{bucket}/{temporary_workspace}", detail=True)
    result = remove_last_modification_time_ms(result)
    expected = {
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir1/file3"): {
            "Key": fsspecfs._strip_protocol(
                f"{bucket}/{temporary_workspace}/dir1/file3"
            ),
            "name": fsspecfs._strip_protocol(
                f"{bucket}/{temporary_workspace}/dir1/file3"
            ),
            "type": "file",
            "size": 0,
        },
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/dir2/file4"): {
            "Key": fsspecfs._strip_protocol(
                f"{bucket}/{temporary_workspace}/dir2/file4"
            ),
            "name": fsspecfs._strip_protocol(
                f"{bucket}/{temporary_workspace}/dir2/file4"
            ),
            "type": "file",
            "size": 0,
        },
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"): {
            "Key": fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"),
            "name": fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1"),
            "type": "file",
            "size": 0,
        },
        fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"): {
            "Key": fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"),
            "name": fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file2"),
            "type": "file",
            "size": 0,
        },
    }
    assert result == expected, f"Expected {expected}, got {result}"

    # Test finding a single file
    result = fsspecfs.find(f"{bucket}/{temporary_workspace}/file1")
    expected = [fsspecfs._strip_protocol(f"{bucket}/{temporary_workspace}/file1")]
    assert result == expected, f"Expected {expected}, got {result}"


def test_du(fsspecfs: Any, bucket: str, temporary_workspace: str):
    dir_path = f"{bucket}/{temporary_workspace}/test_dir"
    nested_dir_path = f"{dir_path}/nested_dir"
    file_path = f"{dir_path}/test_file.txt"
    nested_file_path = f"{nested_dir_path}/nested_file.txt"

    fsspecfs.mkdir(dir_path)
    fsspecfs.mkdir(nested_dir_path)

    with fsspecfs.open(file_path, "wb") as f:
        f.write(b"Hello, World!")
    with fsspecfs.open(nested_file_path, "wb") as f:
        f.write(b"Nested Content")

    # Test total size
    total_size = fsspecfs.du(f"{bucket}/{temporary_workspace}", total=True)
    assert (
        total_size == 13 + 14
    ), f"Expected total size to be {13 + 14}, got {total_size}"

    # Test individual sizes
    sizes = fsspecfs.du(f"{bucket}/{temporary_workspace}", total=False)
    expected_sizes = {
        fsspecfs._strip_protocol(file_path): 13,
        fsspecfs._strip_protocol(nested_file_path): 14,
    }
    assert sizes == expected_sizes, f"Expected sizes {expected_sizes}, got {sizes}"

    with pytest.raises(ValueError, match="maxdepth must be at least 1"):
        fsspecfs.du(f"{bucket}/{temporary_workspace}", total=False, maxdepth=0)

    # Test maxdepth
    sizes_maxdepth_1 = fsspecfs.du(
        f"{bucket}/{temporary_workspace}", total=False, maxdepth=2
    )
    expected_sizes_maxdepth_1 = {fsspecfs._strip_protocol(file_path): 13}
    assert (
        sizes_maxdepth_1 == expected_sizes_maxdepth_1
    ), f"Expected sizes {expected_sizes_maxdepth_1}, got {sizes_maxdepth_1}"


def test_isdir(fsspecfs: Any, bucket: str, temporary_workspace: str):
    # Setup
    dir_name = random_str()
    file_name = random_str()
    dir_path = f"{bucket}/{temporary_workspace}/{dir_name}"
    file_path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    fsspecfs.mkdir(dir_path)
    with fsspecfs.open(file_path, "wb") as f:
        f.write(b"Hello, World!")

    # Test directory
    assert fsspecfs.isdir(dir_path), f"Expected {dir_path} to be a directory"

    # Test file
    assert not fsspecfs.isdir(file_path), f"Expected {file_path} to be a file"

    # Test non-existent path
    non_existent_path = f"{bucket}/{temporary_workspace}/non_existent"
    assert not fsspecfs.isdir(
        non_existent_path
    ), f"Expected {non_existent_path} to not exist"


def test_isfile(fsspecfs: Any, bucket: str, temporary_workspace: str):
    # Setup
    dir_name = random_str()
    file_name = random_str()
    dir_path = f"{bucket}/{temporary_workspace}/{dir_name}"
    file_path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    fsspecfs.mkdir(dir_path)
    with fsspecfs.open(file_path, "wb") as f:
        f.write(b"Hello, World!")

    # Test file
    assert fsspecfs.isfile(file_path), f"Expected {file_path} to be a file"

    # Test directory
    assert not fsspecfs.isfile(dir_path), f"Expected {dir_path} to be a directory"

    # Test non-existent path
    non_existent_path = f"{bucket}/{temporary_workspace}/non_existent"
    assert not fsspecfs.isfile(
        non_existent_path
    ), f"Expected {non_existent_path} to not exist"


def test_rm(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_names = [random_str() for _ in range(5)]
    paths_to_remove = [
        f"{bucket}/{temporary_workspace}/{file_name}.txt" for file_name in file_names
    ]

    for path in paths_to_remove:
        with fsspecfs.open(path, "wb") as f:
            f.write(b"Temporary content")

    fsspecfs.rm(paths_to_remove, recursive=True)

    for path in paths_to_remove:
        assert not fsspecfs.exists(path), f"Path {path} still exists after removal"

    # Remove scheme from paths and test removal again
    paths_without_scheme = [fsspecfs._strip_protocol(path) for path in paths_to_remove]

    for path in paths_without_scheme:
        with fsspecfs.open(path, "wb") as f:
            f.write(b"Temporary content")

    fsspecfs.rm(paths_without_scheme, recursive=True)

    for path in paths_without_scheme:
        assert not fsspecfs.exists(path), f"Path {path} still exists after removal"


def test_cat_file(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path_to_write = f"{bucket}/{temporary_workspace}/{file_name}.bin"
    content = b"Hello, World! This is a test file."
    with fsspecfs.open(path_to_write, "wb") as f:
        f.write(content)

    assert fsspecfs.cat_file(path_to_write) == content, "Failed to read the entire file"

    start = 7
    assert (
        fsspecfs.cat_file(path_to_write, start=start) == content[start:]
    ), "Failed to read from a specific start to the end"

    start, end = 7, 22
    assert (
        fsspecfs.cat_file(path_to_write, start=start, end=end) == content[start:end]
    ), "Failed to read a specific range"

    assert (
        fsspecfs.cat_file(path_to_write, start=-22, end=-7) == content[-22:-7]
    ), "Failed to read with negative indices"


def test_cat(fsspecfs: Any, bucket: str, temporary_workspace: str):
    dir_name = random_str()
    subdir_path = f"{bucket}/{temporary_workspace}/{dir_name}"
    fsspecfs.mkdir(subdir_path)
    file1_path = f"{subdir_path}/file1.txt"
    file2_path = f"{subdir_path}/file2.txt"

    with fsspecfs.open(file1_path, "wb") as f:
        f.write(b"Hello, World!")
    with fsspecfs.open(file2_path, "wb") as f:
        f.write(b"Goodbye, World!")

    # Test Case 1: Single file, recursive=False
    result = fsspecfs.cat(file1_path, recursive=False)
    assert result == b"Hello, World!", "Single file content mismatch"

    # Test Case 2: Directory, recursive=True
    result = fsspecfs.cat(f"{subdir_path}/*", recursive=True)
    expected = {
        fsspecfs._strip_protocol(file1_path): b"Hello, World!",
        fsspecfs._strip_protocol(file2_path): b"Goodbye, World!",
    }
    assert result == expected, "Directory recursive content mismatch"

    # Test Case 3: Error handling with on_error="omit"
    non_existent_path = f"{subdir_path}/nonexistent.txt"
    result = fsspecfs.cat(
        [file1_path, non_existent_path], recursive=True, on_error="omit"
    )
    assert result == {
        fsspecfs._strip_protocol(file1_path): b"Hello, World!"
    }, "Error handling with omit failed"

    result = fsspecfs.cat(f"{subdir_path}/*", recursive=True, on_error="omit")
    assert result == {
        fsspecfs._strip_protocol(file1_path): b"Hello, World!",
        fsspecfs._strip_protocol(file2_path): b"Goodbye, World!",
    }, "Error handling with omit failed"

    # Test Case 4: Error handling with on_error="return"
    result = fsspecfs.cat([file1_path, non_existent_path], on_error="return")
    file1_in_result = fsspecfs._strip_protocol(file1_path) in result
    assert file1_in_result, "file1_path is not in the result"

    non_existent_in_result = isinstance(
        result[fsspecfs._strip_protocol(non_existent_path)], Exception
    )
    assert non_existent_in_result, "non_existent_path is not an Exception in the result"

    fsspecfs.rm(subdir_path, recursive=True)


def test_mv(fsspecfs: Any, bucket: str, temporary_workspace: str):
    source_folder = f"{bucket}/{temporary_workspace}/source_folder"
    target_folder = f"{bucket}/{temporary_workspace}/target_folder"
    test_file_name = "test_file.txt"
    renamed_file_name = "renamed_file.txt"
    test_file_content = b"Hello, ProtonFS!"
    target_file_content = b"Old content"

    fsspecfs.mkdir(source_folder)
    fsspecfs.mkdir(target_folder)
    source_file_path = f"{source_folder}/{test_file_name}"
    target_file_path = f"{target_folder}/{test_file_name}"
    with fsspecfs.open(source_file_path, "wb") as f:
        f.write(test_file_content)

    with fsspecfs.open(target_file_path, "wb") as f:
        f.write(target_file_content)

    fsspecfs.mv(source_file_path, target_file_path)

    assert not fsspecfs.exists(
        source_file_path
    ), "Source file should not exist after move"
    assert fsspecfs.exists(target_file_path), "Target file should exist after move"

    with fsspecfs.open(target_file_path, "rb") as f:
        content = f.read()
        assert (
            content == test_file_content
        ), "Target file content should be overwritten by source file content"

    renamed_file_path = f"{target_folder}/{renamed_file_name}"
    fsspecfs.mv(target_file_path, renamed_file_path)

    assert not fsspecfs.exists(
        target_file_path
    ), "Original file should not exist after rename"
    assert fsspecfs.exists(renamed_file_path), "Renamed file should exist"

    # test mv source dir into target dir recursively
    fsspecfs.rm(source_folder, recursive=True)
    fsspecfs.rm(target_folder, recursive=True)
    fsspecfs.mkdir(source_folder)
    fsspecfs.mkdir(target_folder)
    assert fsspecfs.exists(source_folder)
    assert fsspecfs.exists(target_folder)
    source_file = f"{random_str()}.txt"
    with fsspecfs.open(f"{source_folder}/{source_file}", "wb") as f:
        f.write(test_file_content)
    assert fsspecfs.exists(f"{source_folder}/{source_file}")
    fsspecfs.mv(source_folder, target_folder, recursive=True)
    assert fsspecfs.exists(f"{target_folder}/source_folder")
    assert fsspecfs.exists(f"{target_folder}/source_folder/{source_file}")
    assert not fsspecfs.exists(f"{source_folder}")
    assert not fsspecfs.exists(f"{source_folder}/{source_file}")


def test_get(fsspecfs: Any, bucket: str, temporary_workspace: str):
    remote_file_path = f"{bucket}/{temporary_workspace}/test_get_file.txt"
    local_temp_dir = create_temp_dir()
    local_file_path = f"{local_temp_dir}/test_get_file.txt"
    test_data = b"Data for testing ProtonFileSystem#get method."

    if os.path.exists(local_file_path):
        os.remove(local_file_path)

    with fsspecfs.open(remote_file_path, "wb") as f:
        f.write(test_data)

    fsspecfs.get(remote_file_path, local_file_path)

    assert os.path.exists(local_file_path), "The file was not copied to the local path."
    with open(local_file_path, "rb") as f:
        local_data = f.read()
        assert (
            local_data == test_data
        ), "The data in the local file does not match the expected data."

    os.remove(local_file_path)


def test_put(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = f"{random_str()}.txt"
    with tempfile.TemporaryDirectory() as local_temp_dir:
        local_file_path = f"{local_temp_dir}/{file_name}"
        test_data = b"Data for testing ProtonFileSystem put method."

        with open(local_file_path, "wb") as f:
            f.write(test_data)

        remote_file_path = f"{bucket}/{temporary_workspace}/{file_name}"
        fsspecfs.put(local_file_path, remote_file_path)

        assert fsspecfs.exists(
            remote_file_path
        ), "The file should exist in the remote location after upload."

        downloaded_file_path = f"{local_temp_dir}/downloaded_{file_name}"
        fsspecfs.get(remote_file_path, downloaded_file_path)
        with open(downloaded_file_path, "rb") as f:
            downloaded_data = f.read()
            assert (
                downloaded_data == test_data
            ), "The downloaded file's contents should match the original test data."

        # test lpath and rpath are both dirs
        remote_temp_dir = f"{bucket}/{temporary_workspace}/{random_str()}"

        fsspecfs.mkdir(remote_temp_dir)
        fsspecfs.put(local_temp_dir, remote_temp_dir, recursive=True)
        assert fsspecfs.exists(remote_temp_dir), "The remote directory should exist."
        remote_file_path = os.path.join(
            remote_temp_dir, os.path.basename(local_temp_dir), file_name
        )
        assert fsspecfs.exists(
            remote_file_path
        ), "The remote directory should contain the local directory."

        with fsspecfs.open(remote_file_path, "rb") as f:
            downloaded_data = f.read()
            assert (
                downloaded_data == test_data
            ), "The downloaded file's contents should match the original test data."


def test_put_tree(fsspecfs, bucket, temporary_workspace):
    with tempfile.TemporaryDirectory() as local_temp_dir:
        dir_a = f"{local_temp_dir}/a"
        dir_b = f"{local_temp_dir}/b"
        dir_a_a = f"{dir_a}/a"
        dir_a_a_a = f"{dir_a_a}/a"

        # Create tmp.txt in each directory
        for directory in [dir_a, dir_b, dir_a_a, dir_a_a_a]:
            os.mkdir(directory)
            tmp_file_path = os.path.join(directory, "tmp.txt")
            with open(tmp_file_path, "w") as tmp_file:
                tmp_file.write("Temporary file content")

        remote_temp_dir = f"{bucket}/{temporary_workspace}/{random_str()}"
        fsspecfs.mkdir(remote_temp_dir)
        fsspecfs.put(local_temp_dir, remote_temp_dir, recursive=True)

        # Verify the directory tree on the remote filesystem
        expected_structure = [
            (
                fsspecfs._strip_protocol(f"{remote_temp_dir}"),
                [os.path.basename(local_temp_dir)],
                [],
            ),
            (
                fsspecfs._strip_protocol(
                    f"{remote_temp_dir}/{os.path.basename(local_temp_dir)}"
                ),
                ["a", "b"],
                [],
            ),
            (
                fsspecfs._strip_protocol(
                    f"{remote_temp_dir}/{os.path.basename(local_temp_dir)}/a"
                ),
                ["a"],
                ["tmp.txt"],
            ),
            (
                fsspecfs._strip_protocol(
                    f"{remote_temp_dir}/{os.path.basename(local_temp_dir)}/a/a"
                ),
                ["a"],
                ["tmp.txt"],
            ),
            (
                fsspecfs._strip_protocol(
                    f"{remote_temp_dir}/{os.path.basename(local_temp_dir)}/a/a/a"
                ),
                [],
                ["tmp.txt"],
            ),
            (
                fsspecfs._strip_protocol(
                    f"{remote_temp_dir}/{os.path.basename(local_temp_dir)}/b"
                ),
                [],
                ["tmp.txt"],
            ),
        ]

        actual = list(fsspecfs.walk(remote_temp_dir))
        assert actual == expected_structure


def test_fs_read_block(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data = b"ab\n" + b"a" * (1 * 2**20) + b"\nab"
    with fsspecfs.open(path, "wb") as f:
        f.write(data)

    assert fsspecfs.read_block(path, 0, 3) == b"ab\n"

    non_exist_path = f"{bucket}/{temporary_workspace}/non_exist.txt"
    with pytest.raises(FileNotFoundError):
        fsspecfs.read_block(non_exist_path, 0, 3)


def test_tail(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}.txt"
    data = b"ab\n" + b"a" * (1 * 2**20) + b"\nab"
    with fsspecfs.open(path, "wb") as f:
        f.write(data)

    assert fsspecfs.tail(path, 3) == b"\nab"

    non_exist_path = f"{bucket}/{temporary_workspace}/non_exist.txt"
    with pytest.raises(FileNotFoundError):
        fsspecfs.tail(non_exist_path, 3)


def test_size(fsspecfs: Any, bucket: str, temporary_workspace: str):
    file_name = random_str()
    path = f"{bucket}/{temporary_workspace}/{file_name}"
    content = "hello world"
    with fsspecfs.open(path, "w") as f:
        f.write(content)

    assert fsspecfs.size(path) == len(content)
    assert fsspecfs.size(f"{bucket}/{temporary_workspace}") == 0



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from time import sleep

import pytest

from tosfs.tag import TAGGED_BUCKETS_FILE


@pytest.fixture
def _prepare_tag_env():
    if os.path.exists(TAGGED_BUCKETS_FILE):
        os.remove(TAGGED_BUCKETS_FILE)
    yield
    if os.path.exists(TAGGED_BUCKETS_FILE):
        os.remove(TAGGED_BUCKETS_FILE)


@pytest.mark.usefixtures("_prepare_tag_env")
def test_bucket_tag_action(tosfs, bucket, temporary_workspace):
    tag_mgr = tosfs.bucket_tag_mgr
    if tag_mgr is None:
        return

    tag_mgr.cached_bucket_set = set()
    tag_mgr.add_bucket_tag(bucket)
    sleep(10)
    assert os.path.exists(TAGGED_BUCKETS_FILE)
    with open(TAGGED_BUCKETS_FILE, "r") as f:
        tagged_buckets = f.read()
        assert bucket in tagged_buckets

    assert tag_mgr.bucket_tag_service.get_bucket_tag(bucket)



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from tosfs.core import ENV_NAME_TOSFS_LOGGING_LEVEL, setup_logging


def test_logging_level_debug() -> None:
    # Set the environment variable to DEBUG
    os.environ[ENV_NAME_TOSFS_LOGGING_LEVEL] = "DEBUG"

    # Re-setup logging to apply the new environment variable
    setup_logging()

    # Get the logger and check its level
    logger = logging.getLogger("tosfs")
    assert logger.level == logging.DEBUG



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Generator

import fsspec
import pytest
from fsspec.registry import known_implementations
from tos import EnvCredentialsProvider

from tosfs.core import TosFileSystem, logger
from tosfs.utils import random_str


@pytest.fixture(scope="module")
def _tosfs_env_prepare() -> None:
    if "TOS_ACCESS_KEY" not in os.environ:
        raise EnvironmentError("Can not find TOS_ACCESS_KEY in environment variables.")
    if "TOS_SECRET_KEY" not in os.environ:
        raise EnvironmentError("Can not find TOS_SECRET_KEY in environment variables.")


@pytest.fixture(scope="module")
def tosfs(_tosfs_env_prepare: None) -> TosFileSystem:
    tosfs = TosFileSystem(
        endpoint=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
        max_retry_num=1000,
        connection_timeout=300,
        socket_timeout=300,
        multipart_size=4 << 20,
        multipart_threshold=4 << 20,
    )
    return tosfs


@pytest.fixture(scope="module")
def fsspecfs(_tosfs_env_prepare: None) -> Any:
    known_implementations["tos"] = {"class": "tosfs.TosFileSystem"}

    fsspecfs, _ = fsspec.core.url_to_fs(
        "tos://",
        endpoint_url=os.environ.get("TOS_ENDPOINT"),
        region=os.environ.get("TOS_REGION"),
        credentials_provider=EnvCredentialsProvider(),
    )
    return fsspecfs


@pytest.fixture(scope="module")
def bucket() -> str:
    return os.environ.get("TOS_BUCKET", "proton-ci")


@pytest.fixture(autouse=True)
def temporary_workspace(
    tosfs: TosFileSystem, bucket: str
) -> Generator[str, None, None]:
    workspace = random_str()
    tosfs.mkdirs(f"{bucket}/{workspace}/")
    yield workspace
    try:
        tosfs.rm(f"{bucket}/{workspace}/", recursive=True)
    except Exception as e:
        logger.error(f"Ignore exception : {e}.")
    assert not tosfs.exists(f"{bucket}/{workspace}/")



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import Mock

import pytest
import requests
from tos.exceptions import TosClientError, TosServerError
from tos.http import Response
from urllib3 import HTTPConnectionPool
from urllib3.exceptions import ProtocolError, ReadTimeoutError

from tosfs.retry import _get_sleep_time, is_retryable_exception

mock_resp = Mock(spec=requests.Response)

mock_resp.status_code = 429
mock_resp.headers = {"content-length": "123", "x-tos-request-id": "test-id"}
mock_resp.iter_content = Mock(return_value=[b"chunk1", b"chunk2", b"chunk3"])
mock_resp.json = Mock(return_value={"key": "value"})

response = Response(mock_resp)


@pytest.mark.parametrize(
    ("exception", "expected"),
    [
        (
            TosServerError(
                response,
                "Exceed account external rate limit. Too much throughput in a "
                "short period of time, please slow down.",
                "ExceedAccountExternalRateLimit",
                "KmsJSKDKhjasdlKmsduwRETYHB",
                "",
                "0004-00000001",
            ),
            True,
        ),
        (
            TosClientError(
                "http request timeout",
                ConnectionError(
                    ProtocolError(
                        "Connection aborted.",
                        ConnectionResetError(104, "Connection reset by peer"),
                    )
                ),
            ),
            True,
        ),
        (
            TosClientError(
                "{'message': 'http request timeout', "
                "'case': \"('Connection aborted.', "
                "ConnectionResetError(104, 'Connection reset by peer'))\", "
                "'request_url': "
                "'http://proton-ci.tos-cn-beijing.volces.com/"
                "nHnbR/yAlen'}",
                TosClientError(
                    "http request timeout",
                    ConnectionError(
                        ProtocolError(
                            "Connection aborted.",
                            ConnectionResetError(104, "Connection reset by peer"),
                        )
                    ),
                ),
            ),
            True,
        ),
        (
            requests.exceptions.ConnectionError(
                HTTPConnectionPool(host="proton-ci.tos-cn-beijing.volces.com", port=80)
            ),
            True,
        ),
        (
            ReadTimeoutError(None, message="", url=""),
            True,
        ),
    ],
)
def test_is_retry_exception(
    exception,
    expected,
):
    assert is_retryable_exception(exception) == expected


def test_get_sleep_time():
    mock_resp = Mock(spec=requests.Response)

    sleep_time = 1000

    mock_resp.status_code = 429
    mock_resp.headers = {
        "Content-Type": "application/json",
        "Date": "Thu, 07 Nov 2024 17:19:56 GMT",
        "Retry-After": str(sleep_time),
        "Server": "TosServer",
        "x-tos-ec": "0004-00000007",
        "x-tos-id-2": "0000-00000-0000",
        "x-tos-request-id": "0000-0000-0000",
        "x-tos-server-time": "1",
        "Content-Length": "277",
    }
    mock_resp.iter_content = Mock(return_value=[b"chunk1", b"chunk2", b"chunk3"])
    mock_resp.json = Mock(return_value={"key": "value"})

    response = Response(mock_resp)

    err = TosServerError(
        response,
        "Exceed account external rate limit. Too much throughput in a "
        "short period of time, please slow down.",
        "ExceedAccountExternalRateLimit",
        "KmsJSKDKhjasdlKmsduwRETYHB",
        "",
        "0004-00000001",
    )

    assert _get_sleep_time(err, 1) == sleep_time



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module contains utility functions for the tosfs package."""
import random
import re
import string
import tempfile
from typing import Generator, Tuple


def random_str(length: int = 5) -> str:
    """Generate a random string of the given length.

    Args:
    ----
    length (int): The length of the random string.

    Returns:
    -------
    str: The random string.

    """
    return "".join(random.choices(string.ascii_letters, k=length))


def create_temp_dir() -> str:
    """Create a temporary directory for testing purposes.

    Returns
    -------
    str: The path of the created temporary directory.

    """
    return tempfile.mkdtemp()


def find_bucket_key(tos_path: str) -> Tuple[str, str]:
    """It's a helper function to find bucket and key pair.

    It receives a tos path such that the path
    is of the form: bucket/key.
    It will return the bucket and the key represented by the tos path.
    """
    bucket_format_list = [
        re.compile(
            r"^(?P<bucket>:tos:[a-z\-0-9]*:[0-9]{12}:accesspoint[:/][^/]+)/?"
            r"(?P<key>.*)$"
        ),
        re.compile(
            r"^(?P<bucket>:tos-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:]"
            # pylint: disable=line-too-long
            r"[a-zA-Z0-9\-]{1,63}[/:](bucket|accesspoint)[/:][a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
        ),
        re.compile(
            r"^(?P<bucket>:tos-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:]"
            r"[a-zA-Z0-9\-]{1,63}[/:]bucket[/:]"
            r"[a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
        ),
        re.compile(
            r"^(?P<bucket>:tos-object-lambda:[a-z\-0-9]+:[0-9]{12}:"
            r"accesspoint[/:][a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
        ),
        re.compile(r"^tos://(?P<bucket>[^/]+)/(?P<key>.*)$"),
    ]
    for bucket_format in bucket_format_list:
        match = bucket_format.match(tos_path)
        if match:
            return match.group("bucket"), match.group("key")
    tos_components = tos_path.split("/", 1)
    bucket = tos_components[0]
    tos_key = ""
    if len(tos_components) > 1:
        tos_key = tos_components[1]
    return bucket, tos_key


def get_brange(size: int, block: int) -> Generator[Tuple[int, int], None, None]:
    """Chunk up a file into zero-based byte ranges.

    Parameters
    ----------
    size : int
        file size
    block : int
        block size

    """
    for offset in range(0, size, block):
        yield offset, min(offset + block - 1, size - 1)



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""It contains everything about certification via a file-based provider."""

import threading
from datetime import datetime, timedelta
from json import JSONDecodeError
from typing import Optional
from xml.etree import ElementTree

import requests
from tos.consts import ECS_DATE_FORMAT
from tos.credential import Credentials, CredentialsProvider

from tosfs.core import logger
from tosfs.exceptions import TosfsCertificationError

CERTIFICATION_REFRESH_INTERVAL_MINUTES = 60
CERTIFICATION_MAX_VALID_PERIOD_HOURS = 12


class FileCredentialsProvider(CredentialsProvider):
    """The class provides the credentials from a file.

    The file should be in the format of:
      <configuration>
        <property>
            <name>fs.tos.access-key-id</name>
            <value>access_key</value>
        </property>
        <property>
            <name>fs.tos.secret-access-key</name>
            <value>secret_key</value>
        </property>
        <property>
            <name>fs.tos.session-token</name>
            <value>session_token</value>
        </property>
      </configuration>

    It can only receive a file path which exists in the local file system.

    Note :
    1. The default refresh interval is 60 minutes.
    2. The maximum valid period for provisional certification is 12 hours.

    This provider will cache the credentials and refresh them every 60 minutes.
    And note that, it only reads the credentials from the file and refreshes itself,
    to guarantee the credentials are always up-to-date,
    we need a service update it internally.

    Examples
    --------
    >>> from tosfs import TosFileSystem
    >>> tosfs = TosFileSystem(
    >>>             endpoint="tos-cn-beijing.volcengine.com",
    >>>             regions="cn-bejing",
    >>>             credentials_provider=FileCredentialsProvider("dummy_path"))
    >>> tosfs.ls("tos://bucket/path")

    """

    def __init__(
        self,
        file_path: str,
        refresh_interval_min: int = CERTIFICATION_REFRESH_INTERVAL_MINUTES,
    ) -> None:
        """Initialize the FileCredentialsProvider."""
        self.file_path = file_path
        self.refresh_interval = refresh_interval_min
        if self.refresh_interval <= 0:
            logger.warning(
                f"Invalid refresh interval {self.refresh_interval}, "
                f"set to default value: 60 minutes"
            )
            self.refresh_interval = CERTIFICATION_REFRESH_INTERVAL_MINUTES
        if self.refresh_interval > CERTIFICATION_MAX_VALID_PERIOD_HOURS * 60:
            logger.warning(
                f"Invalid refresh interval {self.refresh_interval}, "
                f"set to maximum value: 60 minutes"
            )
            self.refresh_interval = CERTIFICATION_REFRESH_INTERVAL_MINUTES
        self.prev_refresh_time: Optional[datetime] = None
        self.credentials = None
        self._lock = threading.Lock()

    def get_credentials(self) -> Credentials:
        """Get the credentials from the file.

        Returns
        -------
        Credentials: The credentials object.

        Raises
        ------
        TosfsCertificationError: If the credentials cannot be retrieved.

        """
        res = self._try_get_credentials()
        if res is not None:
            return res
        with self._lock:
            try:
                res = self._try_get_credentials()
                if res is not None:
                    return res

                with open(self.file_path, "r") as f:
                    logger.debug(
                        f"Trying to refresh the credentials from file: "
                        f"{self.file_path}"
                    )
                    tree = ElementTree.parse(f)  # noqa S314
                    root = tree.getroot()

                    access_key_element = root.find(
                        ".//property[name='fs.tos.access-key-id']/value"
                    )
                    secret_key_element = root.find(
                        ".//property[name='fs.tos.secret-access-key']/value"
                    )
                    session_token_element = root.find(
                        ".//property[name='fs.tos.session-token']/value"
                    )

                    access_key = (
                        access_key_element.text
                        if access_key_element is not None
                        else None
                    )
                    secret_key = (
                        secret_key_element.text
                        if secret_key_element is not None
                        else None
                    )
                    session_token = (
                        session_token_element.text
                        if session_token_element is not None
                        else None
                    )

                    if None in (
                        access_key,
                        secret_key,
                        session_token,
                    ):
                        raise TosfsCertificationError(
                            "Missing required credential elements in the file"
                        )

                    self.prev_refresh_time = datetime.now()
                    self.credentials = Credentials(
                        access_key, secret_key, session_token
                    )
                    logger.debug(
                        f"Successfully refreshed the credentials from file: "
                        f"{self.file_path}"
                    )

                return self.credentials
            except Exception as e:
                raise TosfsCertificationError("Get certification error: ") from e

    def _try_get_credentials(self) -> Optional[Credentials]:
        if self.prev_refresh_time is None or self.credentials is None:
            return None
        if (
            datetime.now() - self.prev_refresh_time
        ).total_seconds() / 60 > CERTIFICATION_REFRESH_INTERVAL_MINUTES:
            logger.debug(
                f"Credentials are expired, "
                f"will try to refresh the credentials from file: "
                f"{self.file_path}"
            )
            return None
        return self.credentials


class UrlCredentialsProvider(CredentialsProvider):
    """The class provides the credentials from an url."""

    def __init__(self, credential_url: str):
        """Initialize the UrlCredentialsProvider."""
        if not credential_url:
            raise TosfsCertificationError("The credential_url param must not be empty.")
        self._lock = threading.Lock()
        self.expires: Optional[datetime] = None
        self.credentials = None
        self.credential_url = credential_url

    def get_credentials(self) -> Credentials:
        """Get the credentials from the url."""
        res = self._try_get_credentials()
        if res is not None:
            return res
        with self._lock:
            try:
                res = self._try_get_credentials()
                if res is not None:
                    return res

                from tosfs.retry import retryable_func_executor

                res = retryable_func_executor(
                    requests.get,
                    args=(self.credential_url,),
                    kwargs={"timeout": 60},
                    max_retry_num=3,
                )
                res_body = res.json()
                self.credentials = Credentials(
                    res_body.get("AccessKeyId"),
                    res_body.get("SecretAccessKey"),
                    res_body.get("SessionToken"),
                )
                self.expires = datetime.strptime(
                    res_body.get("ExpiredTime"), ECS_DATE_FORMAT
                )
                return self.credentials
            except JSONDecodeError as e:
                logger.error(
                    f"Failed to parse the response from the url: "
                    f"{self.credential_url}, error: {e}, detail: {e.doc}"
                )
                if self.expires is not None and (
                    datetime.now().timestamp() < self.expires.timestamp()
                ):
                    return self.credentials
                raise TosfsCertificationError(
                    "Get token failed, please check the credential url"
                ) from e
            except Exception as e:
                if self.expires is not None and (
                    datetime.now().timestamp() < self.expires.timestamp()
                ):
                    return self.credentials
                raise TosfsCertificationError(f"Get token failed: {e}") from e

    def _try_get_credentials(self) -> Optional[Credentials]:
        if self.expires is None or self.credentials is None:
            return None
        if (
            datetime.now().timestamp()
            > (self.expires - timedelta(minutes=10)).timestamp()
        ):
            return None
        return self.credentials


class NoLockUrlCredentialsProvider(CredentialsProvider):
    """The class provides the credentials from an url.

    It does not use lock to protect the credentials.
    Due to threading.Lock is not serializable,
    it can not be used in Ray remote function.
    """

    def __init__(self, credential_url: str):
        """Initialize the UrlCredentialsProvider."""
        if not credential_url:
            raise TosfsCertificationError("The credential_url param must not be empty.")
        self.expires: Optional[datetime] = None
        self.credentials = None
        self.credential_url = credential_url

    def get_credentials(self) -> Credentials:
        """Get the credentials from the url."""
        res = self._try_get_credentials()
        if res is not None:
            return res
        try:
            res = self._try_get_credentials()
            if res is not None:
                return res

            from tosfs.retry import retryable_func_executor

            res = retryable_func_executor(
                requests.get,
                args=(self.credential_url,),
                kwargs={"timeout": 60},
                max_retry_num=3,
            )
            res_body = res.json()
            self.credentials = Credentials(
                res_body.get("AccessKeyId"),
                res_body.get("SecretAccessKey"),
                res_body.get("SessionToken"),
            )
            self.expires = datetime.strptime(
                res_body.get("ExpiredTime"), ECS_DATE_FORMAT
            )
            return self.credentials
        except JSONDecodeError as e:
            logger.error(
                f"Failed to parse the response from the url: {self.credential_url}, "
                f"error: {e}, detail: {e.doc}"
            )
            if self.expires is not None and (
                datetime.now().timestamp() < self.expires.timestamp()
            ):
                return self.credentials
            raise TosfsCertificationError(
                "Get token failed, please check the credential url"
            ) from e
        except Exception as e:
            if self.expires is not None and (
                datetime.now().timestamp() < self.expires.timestamp()
            ):
                return self.credentials
            raise TosfsCertificationError(f"Get token failed: {e}") from e

    def _try_get_credentials(self) -> Optional[Credentials]:
        if self.expires is None or self.credentials is None:
            return None
        if (
            datetime.now().timestamp()
            > (self.expires - timedelta(minutes=10)).timestamp()
        ):
            return None
        return self.credentials



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta
from unittest.mock import mock_open, patch

import pytest

from tosfs.certification import FileCredentialsProvider
from tosfs.exceptions import TosfsCertificationError


@patch(
    "builtins.open",
    new_callable=mock_open,
    read_data="""
<configuration>
    <property>
        <name>fs.tos.access-key-id</name>
        <value>access_key</value>
    </property>
    <property>
        <name>fs.tos.secret-access-key</name>
        <value>secret_key</value>
    </property>
    <property>
        <name>fs.tos.session-token</name>
        <value>session_token</value>
    </property>
</configuration>
""",
)
def test_get_credentials(mock_file):
    provider = FileCredentialsProvider("dummy_path")
    credentials = provider.get_credentials()
    assert credentials.access_key_id == "access_key"
    assert credentials.access_key_secret == "secret_key"  # noqa S105
    assert credentials.security_token == "session_token"  # noqa S105


@patch(
    "builtins.open",
    new_callable=mock_open,
    read_data="""
<configuration>
    <property>
        <name>fs.tos.access-key-id</name>
        <value>access_key</value>
    </property>
    <property>
        <name>fs.tos.secret-access-key</name>
        <value>secret_key</value>
    </property>
    <property>
        <name>fs.tos.session-token</name>
        <value>session_token</value>
    </property>
</configuration>
""",
)
def test_get_credentials_with_refresh(mock_file):
    provider = FileCredentialsProvider("dummy_path")
    provider.prev_refresh_time = datetime.now() - timedelta(minutes=61)
    credentials = provider.get_credentials()
    assert credentials.access_key_id == "access_key"
    assert credentials.access_key_secret == "secret_key"  # noqa S105
    assert credentials.security_token == "session_token"  # noqa S105


@patch(
    "builtins.open",
    new_callable=mock_open,
    read_data="""
<configuration>
    <property>
        <name>fs.tos.access-key-id</name>
        <value>access_key</value>
    </property>
    <property>
        <name>fs.tos.secret-access-key</name>
        <value>secret_key</value>
    </property>
    <property>
        <name>fs.tos.session-token</name>
        <value>session_token</value>
    </property>
</configuration>
""",
)
def test_get_credentials_error(mock_file):
    provider = FileCredentialsProvider("dummy_path")
    with patch("xml.etree.ElementTree.parse", side_effect=Exception("Parse error")):
        with pytest.raises(TosfsCertificationError):
            provider.get_credentials()


@patch(
    "builtins.open",
    new_callable=mock_open,
    read_data="""
<configuration>
    <property>
        <key>fs.tos.access-key-id</name>
        <value>access_key</value>
    </property>
    <property>
        <key>fs.tos.secret-access-key</name>
        <value>secret_key</value>
    </property>
    <property>
        <key>fs.tos.session-token</name>
        <value>session_token</value>
    </property>
</configuration>
""",
)
def test_wrong_file_format_error(mock_file):
    provider = FileCredentialsProvider("dummy_path")
    provider.prev_refresh_time = datetime.now() - timedelta(minutes=61)
    with pytest.raises(TosfsCertificationError):
        provider.get_credentials()


@patch(
    "builtins.open",
    new_callable=mock_open,
    read_data="""
<configuration>
    <property>
        <name>fs.tos.access-key-id</name>
        <value>access_key</value>
    </property>
    <property>
        <name>fs.tos.secret-access-key</name>
        <value>secret_key</value>
    </property>
    <property>
        <name>fs.tos.session-token</name>
        <value>session_token</value>
    </property>
</configuration>
""",
)
def test_no_refresh_within_interval(mock_file):
    provider = FileCredentialsProvider("dummy_path")
    provider.get_credentials()

    provider.prev_refresh_time = datetime.now() - timedelta(minutes=30)

    mock_file().read_data = """
    <configuration>
        <property>
            <name>fs.tos.access-key-id</name>
            <value>new_access_key</value>
        </property>
        <property>
            <name>fs.tos.secret-access-key</name>
            <value>new_secret_key</value>
        </property>
        <property>
            <name>fs.tos.session-token</name>
            <value>new_session_token</value>
        </property>
    </configuration>
    """

    new_credentials = provider.get_credentials()

    assert new_credentials.access_key_id == "access_key"
    assert new_credentials.access_key_secret == "secret_key"  # noqa S105
    assert new_credentials.security_token == "session_token"  # noqa S105



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""It contains exceptions definition for the tosfs package."""
from typing import Optional


class TosfsError(Exception):
    """Base class for all tosfs exceptions."""

    def __init__(self, msg: str, cause: Optional[Exception] = None):
        """Initialize the base class for all exceptions in the tosfs package."""
        super().__init__(msg, cause)
        self.message = msg
        self.cause = cause

    def __str__(self) -> str:
        """Return the string representation of the exception."""
        error = {"message": self.message, "case": str(self.cause)}
        return str(error)


class TosfsCertificationError(TosfsError):
    """Exception class for certification related exception."""

    def __init__(self, message: str, cause: Optional[Exception] = None):
        """Initialize the exception class for certification related exception."""
        super().__init__(message, cause)
